/**************************************************************************************
Author:         David Szybala
Website:        www.davidszybala.com
Alt Website:    www.twistedgerbil.com
Program:        bitmap_cleaner.ms
Date:           11/5/14
Description:    This scripts provides a number of methods for altering bitmap paths.
***************************************************************************************/

try (bitmap_cleaner_form.close()) catch()


struct BitmapCleaner
(
    bitmap_nodes = #(),
    bitmap_paths = #(),
    search_options = dotnetClass "System.IO.SearchOption",
    search_engine = dotnetClass "System.IO.Directory",

    /**************************************************************************
    Returns a list of all bitmap textures. Supports VRayHDRI textures
    **************************************************************************/
    fn get_bitmap_nodes =
    (
        local all_bitmaps = getClassInstances BitmapTexture

        if VRayHDRI != undefined do
            join all_bitmaps (getClassInstances VRayHDRI)

        all_bitmaps --return value
    ), --end fn
    
    /**************************************************************************
    CRITICAL - BUG FOUND! Found a bitmap node with undefined assigned to
    filename (which will error the sorting method). No clue how that is even
    possible, but it happened. This function will assign "" to any map with
    undefined as the filename
    **************************************************************************/
    fn validate_bitmap_path bitmap_node =
    (
        case of
        (
            (isProperty bitmap_node "filename"): if bitmap_node.filename == undefined do bitmap_node.filename = ""
            (isProperty bitmap_node "HDRIMapName"): if bitmap_node.HDRIMapName == undefined do bitmap_node.HDRIMapName = ""
        ) --end case
    ), --end fn

    /**************************************************************************
    Returns the file path for the current bitmap node. Supports VRayHDRI textures
    **************************************************************************/
    fn get_node_path bitmap_node =
    (
        local file_path = ""

        case of
        (
            (isProperty bitmap_node "filename"): file_path = bitmap_node.filename
            (isProperty bitmap_node "HDRIMapName"): file_path = bitmap_node.HDRIMapName
        ) --end case
        
        file_path --return value
    ), --end fn

    /**************************************************************************
    Sorts a list by the objects file path
    **************************************************************************/
    fn sort_by_file_path bitmap_node_one bitmap_node_two =
    (
        local name_one = ""
        local name_two = ""

        path_one = this.get_node_path bitmap_node_one
        path_two = this.get_node_path bitmap_node_two

        case of
        (
            (path_one < path_two): -1
            (path_one > path_two): 1
            default: 0
        ) --end case
    ), --end fn
    
    /**************************************************************************
    Collects the file paths for all found bitmap nodes
    **************************************************************************/
    fn collect_bitmap_paths =
    (
        this.bitmap_paths = for node in this.bitmap_nodes collect this.get_node_path node
    ), --end fn

    /**************************************************************************
    Collects all bitmap nodes and their paths in the current scene
    **************************************************************************/
    fn collect_bitmap_nodes =
    (
        this.bitmap_nodes = this.get_bitmap_nodes()
        
        --CRITICAL! Needed for bug fix!
        for current in this.bitmap_nodes do
            this.validate_bitmap_path current
        
        qsort this.bitmap_nodes this.sort_by_file_path
        this.collect_bitmap_paths()
    ), --end fn
    
    /**************************************************************************
    Searches 'search_directory' for the file name from 'file_path' 
    **************************************************************************/
    fn search_for_file search_directory file_path include_subs =
    (
        local file_name = filenameFromPath file_path
        
        --CRITICAL! Searching a directory without the correct permissions will
        --throw an error!
        try
        (
            if include_subs then
                result = this.search_engine.GetFiles search_directory file_name this.search_options.AllDirectories
            else
                result = this.search_engine.GetFiles search_directory file_name this.search_options.TopDirectoryOnly
        ) --end try
        catch
        (
            result = #()
        ) --end catch
        
        result --return value
    ), --end fn

    /**************************************************************************
    Returns a complete path by inserting 'new_section' into 'current_path' at
    the position based on 'assemble_mode' (path/name/ext/all)
    **************************************************************************/
    fn get_complete_path current_path new_section assemble_mode =
    (
        local complete_path = ""

        case assemble_mode of
        (
            (1): complete_path = new_section + "/" + (filenameFromPath current_path)
            (2): complete_path = (getFilenamePath current_path) + new_section + (getFilenameType current_path)
            (3): complete_path = (getFilenamePath current_path) + (getFilenameFile current_path) + "." + new_section
            (4): complete_path = new_section
        ) --end case

        if complete_path != "" do
            complete_path = pathconfig.normalizePath complete_path

        complete_path --return value
    ), --end fn
    
    /**************************************************************************
    Renames the 'source_path' to the 'destination_path' Will also make
    directories if the destination is in another location. Returns true on
    success
    **************************************************************************/
    fn rename_file source_path destination_path =
    (
        local dest_path = getFilenamePath destination_path
        local result = true
        
        if source_path != destination_path do
        (
            if not (doesFileExist dest_path) do
                result = makeDir dest_path all:true
            
            if result do
                result = renameFile source_path destination_path
        ) --end if
        
        result --return value
    ), --end fn

    /**************************************************************************
    Sets all bitmap nodes with the file path 'from_path' to the new path
    'to_path'
    **************************************************************************/
    fn set_bitmap_path from_path to_path =
    (
        for i = 1 to this.bitmap_paths.count do
        (
            if this.bitmap_paths[i] == from_path do
            (
                case of
                (
                    (isProperty this.bitmap_nodes[i] "filename"): this.bitmap_nodes[i].filename = to_path
                    (isProperty this.bitmap_nodes[i] "HDRIMapName"): this.bitmap_nodes[i].HDRIMapName = to_path
                ) --end case
            ) --end if
        ) --end for
    ), --end fn

    /**************************************************************************
    Returns the path part of the current path without the last backslash
    **************************************************************************/
    fn get_file_path current_path =
    (
        local file_path = getFilenamePath current_path

        if file_path[file_path.count] == "\\" or file_path[file_path.count] == "/" do
            file_path = substring file_path 1 (file_path.count - 1)

        file_path --return value
    ), --end fn

    /**************************************************************************
    Returns the extension of the current path without the period
    **************************************************************************/
    fn get_file_extension current_path =
    (
        local file_extension = getFilenameType current_path

        if file_extension[1] == "." do
            file_extension = substring file_extension 2 (file_extension.count)

        file_extension --return value
    ), --end fn
    
    /**************************************************************************
    Returns a part of the 'current_path' based on the mode (path/name/ext)
    **************************************************************************/
    fn get_path_section current_path section_mode =
    (
        local path_section = ""

        case section_mode of
        (
            (1): path_section = this.get_file_path current_path
            (2): path_section = getFilenameFile current_path
            (3): path_section = this.get_file_extension current_path
            (4): path_section = current_path
        ) --end case

        path_section --return value
    ), --end fn

    /**************************************************************************
    Returns the result of a case sensitive find and replace operation on a
    file path
    **************************************************************************/
    fn get_case_replace_path file_path find_text replace_text first_only =
    (
        local new_path = file_path
        local index = 1
        local continue_search = true

        if (matchPattern new_path pattern:("*" + find_text + "*") ignorecase:false) do
        (
            if first_only then
            (
                while continue_search do
                (
                    if new_path[index] == find_text[1] do
                    (
                        if (substring new_path index find_text.count) == find_text do
                        (
                            new_path = replace new_path index find_text.count replace_text
                            continue_search = false
                        ) --end if
                    ) --end if

                    index += 1
                ) --end while
            ) --end if
            else
                new_path = substituteString file_path find_text replace_text
        ) --end if

        new_path --return value
    ), --end fn

    /**************************************************************************
    Returns the result of a case insensitive find and replace operation on a
    file path
    **************************************************************************/
    fn get_replace_path file_path find_text replace_text first_only =
    (
        local new_path = file_path
        local index = 1

        if (matchPattern new_path pattern:("*" + find_text + "*") ignorecase:true) do
        (
            index = findString new_path find_text

            while index != undefined do
            (
                new_path = replace new_path index find_text.count replace_text

                if first_only then
                    index = undefined
                else
                    index = findString new_path find_text
            ) --end while
        ) --end if

        new_path --return value
    ), --end fn

    /**************************************************************************
    Performs a find/replace on 'current_path' and returns the result
    **************************************************************************/
    fn get_find_replace_result current_path find_mode find_text replace_text match_case first_only =
    (
        local new_path = this.get_path_section current_path find_mode

        if match_case then
            new_path = this.get_case_replace_path new_path find_text replace_text first_only
        else
            new_path = this.get_replace_path new_path find_text replace_text first_only

        new_path = this.get_complete_path current_path new_path find_mode
        
        new_path --return value
    ), --end fn

    /**************************************************************************
    Returns an absolute path for the argument "file_path" NOTE: If the path
    is not already an absolute path, map paths are used first
    (since they will return null if the file is not found), then the project
    directory is used (since it returns a value no matter what)
    **************************************************************************/
    fn get_absolute_path file_path =
    (
        local absolute_path = pathConfig.normalizePath file_path

        --CRITICAL! Max needs relative paths to start with a '.' to recognize them!
        if (pathConfig.isPathRootedAtBackslash absolute_path) do
        (
            if not (matchPattern absolute_path pattern:"\\\\*") do
                absolute_path = "." + absolute_path
        ) --end if

        --CRITICAL! Store cleaned copy of file path!
        file_path = absolute_path

        absolute_path = mapPaths.getFullFilePath file_path

        if absolute_path == "" do
            absolute_path = pathConfig.convertPathToAbsolute file_path

        absolute_path --return value
    ), --end fn

    /**************************************************************************
    Returns a list of all maps paths that do not exist
    **************************************************************************/
    fn get_missing_maps =
    (
        local map_paths = makeUniqueArray this.bitmap_paths
        local missing_maps = #()

        for current_path in map_paths do
        (
            search_path = this.get_absolute_path current_path

            if not (doesFileExist search_path) do
                append missing_maps current_path
        ) --end for

        missing_maps --return value
    ), --end fn
    
    /**************************************************************************
    Returns all bitmap nodes using the 'current_path'
    **************************************************************************/
    fn get_nodes_using_path current_path =
    (
        local return_nodes = #()

        for i = 1 to this.bitmap_paths.count do
        (
            if current_path == this.bitmap_paths[i] do
                append return_nodes this.bitmap_nodes[i]
        ) --end for

        return_nodes --return value
    ), --end fn

    /**************************************************************************
    Exports the map 'current_path' to a new directory. Can also update the
    scene path
    **************************************************************************/
    fn export_maps export_path current_path update_scene =
    (
        local map_paths = makeUniqueArray this.bitmap_paths
        local export_success = true

        source_path = this.get_absolute_path current_path
        destination_path = export_path + "/" + (filenameFromPath current_path)
        destination_path = pathconfig.normalizePath destination_path
        
        if source_path == destination_path do
            return true

        if (doesFileExist destination_path) do
            deleteFile destination_path

        if (copyFile source_path destination_path) then
        (
            if update_scene do
                this.set_bitmap_path current_path destination_path
        ) --end if
        else
        (
            print("ERROR: Could not copy file " + source_path + " to " + destination_path)
            export_success = false
        ) --end else

        export_success --return value
    ), --end fn

    /**************************************************************************
    Returns all root materials/textures using 'bitmap_node'
    **************************************************************************/
    fn get_materials_using_node bitmap_node =
    (
        local material_dependents = #()

        for current_node in (refs.dependents bitmap_node) \
            where (superClassOf current_node) == material or (superClassOf current_node) == textureMap do
                append material_dependents current_node

        material_dependents --return value
    ), --end fn
    
    /**************************************************************************
    Resets all unwrap modifier textures
    **************************************************************************/
    fn reset_unwrap_textures =
    (
        for current in (getClassInstances Unwrap_UVW) do
        (
            current.baseMaterial_list.count = 1
            current.baseMaterial = StandardMaterial name:"Unwrap_Material"
            current.texMapList.count = 1
            current.texMapIDList.count = 1
            current.texMapList[1] = Checker()
            current.texMapIDList[1] = -1
        ) --end for
    ) --end fn
) --end struct


( --local scope
    global bitmap_cleaner_form
    local bitmap_cleaner_ini = getDir #plugcfg + "\\BitmapCleaner.ini"
    
    --CRITICAL! These controls/variables need to be accessed in event handlers!
    local cleaner = BitmapCleaner()
    
    local search_directory_text = undefined
    local include_sub_directory_checkbox = undefined
    
    local mode_path_radio = undefined
    local mode_filename_radio = undefined
    local mode_extension_radio = undefined
    local mode_all_radio = undefined
    local option_affect_files_checkbox = undefined
    
    local modify_text = undefined
    
    local find_text = undefined
    local replace_text = undefined
    local match_case_checkbox = undefined
    local first_only_checkbox = undefined
    
    local bitmap_group = undefined
    local bitmap_list = undefined
    local progress_bar = undefined
    local status_label = undefined
    
    /********************************************************************
    Form closing event handler. Stores form information for next time
    ********************************************************************/
    fn bitmap_cleaner_form_closing sender events =
    (
        local form_location = bitmap_cleaner_form.location
        local form_size = bitmap_cleaner_form.size

        form_location = point2 form_location.x form_location.y
        form_size = point2 form_size.width form_size.height

        setIniSetting bitmap_cleaner_ini "Dialog" "Position" (form_location as string)
        setIniSetting bitmap_cleaner_ini "Dialog" "Size" (form_size as string)

        bitmap_cleaner_form.Dispose()
        bitmap_cleaner_form = undefined
    ) --end fn

    /**********************************************************************
    Updates the bitmap path list box. Returns the total number of unique
    maps found
    **********************************************************************/
    fn refresh_bitmap_list =
    (
        local bitmap_paths = makeUniqueArray cleaner.bitmap_paths
        
        bitmap_list.BeginUpdate()
        
        for i = (bitmap_list.Items.Count - 1) to 0 by -1 do
            bitmap_list.Items.Remove (bitmap_list.Items.Item i)
        
        for current_path in bitmap_paths do
        (
            search_path = cleaner.get_absolute_path current_path
            
            current_item = dotNetObject "System.Windows.Forms.ListViewItem" current_path
            current_item.SubItems.Add (search_path.count as string)

            case of
            (
                (not (doesFileExist search_path)):
                    (
                        current_item.SubItems.Add ("ERROR: File does not exist " + search_path)
                        current_item.ForeColor = (dotNetClass "System.Drawing.Color").FromArgb 255 125 0 0
                    ) --end case 1
                    
                (search_path.count > 255):
                    (
                        current_item.SubItems.Add ("WARNING: Path length exceeds 255 characters " + search_path)
                        current_item.ForeColor = (dotNetClass "System.Drawing.Color").FromArgb 255 125 125 0
                    ) --end case 2
                default:
                    (
                        current_item.SubItems.Add "OK"
                        sys_color = (colorMan.getColor #text) * 255
                        current_item.ForeColor = (dotNetClass "System.Drawing.Color").FromArgb 255 sys_color[1] sys_color[2] sys_color[3]
                    ) --end default
            ) --end case
            
            bitmap_list.Items.Add current_item
        )--end for
        
        bitmap_list.EndUpdate()
        
        bitmap_paths.count --return value
    ) --end fn
    
    /**********************************************************************
    Search browse button clicked event handler. Allows the user to browse for a directory
    **********************************************************************/
    fn search_browse_button_pressed sender events =
    (
        local new_path = getSavePath caption:"Search Directory"

        if new_path != undefined do
            search_directory_text.text = new_path
    ) --end fn
    
    /**********************************************************************
    Search directory button clicked event handler. Searches the selected directory for the
    selected bitmaps
    **********************************************************************/
    fn search_directory_button_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text
        local error_counter = 0
        
        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No maps selected"
            return false
        ) --end if
        
        if search_directory_text.text == "" do
        (
            status_label.text = "ERROR: Search directory is blank"
            return false
        ) --end if
        
        if not (doesFileExist search_directory_text.text) do
        (
            status_label.text = "ERROR: Search directory does not exist"
            return false
        ) --end if
        
        progress_bar.Value = 0
        progress_bar.Visible = true
        progress_bar.Maximum = bitmap_paths.count
        
        status_label.text = "INFO: Searching for maps. Please wait..."
        if ((maxVersion())[1] / 1000) > 12 do windows.processPostedMessages()
        
        for current in bitmap_paths do
        (
            result = cleaner.search_for_file search_directory_text.text current include_sub_directory_checkbox.checked
            
            if result.count > 0 then
                cleaner.set_bitmap_path current result[1]
            else
            (
                print("WARNING: Could not find bitmap " + current)
                error_counter += 1
            ) --end else
            
            progress_bar.PerformStep()
        ) --end for
        
        if error_counter < bitmap_paths.count do
        (
            cleaner.collect_bitmap_paths()
            refresh_bitmap_list()
        ) --end if
        
        progress_bar.Visible = false
        progress_bar.Value = 0

        if error_counter == 0 then
            status_label.text = "INFO: Search complete. All maps found"
        else
            status_label.text = "WARNING: Search complete. Some maps not found. Check listener for details"
    ) --end fn

    /**********************************************************************
    Modify paths button clicked event handler. Modifies the path/filename/ext
    of the selected bitmaps
    **********************************************************************/
    fn modify_path_button_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text
        local modify_mode = 1
        local result = true
        local error_counter = 0

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No maps selected"
            return false
        ) --end if

        if modify_text.text == "" do
        (
            status_label.text = "ERROR: Modify field is blank"
            return false
        ) --end if
        
        case of
        (
            (mode_filename_radio.checked): modify_mode = 2
            (mode_extension_radio.checked): modify_mode = 3
            (mode_all_radio.checked): modify_mode = 4
        ) --end case
        
        status_label.text = "INFO: Modifying selected maps. Please wait..."
        if ((maxVersion())[1] / 1000) > 12 do windows.processPostedMessages()

        for current_path in bitmap_paths do
        (
            new_path = cleaner.get_complete_path current_path modify_text.text modify_mode
            
            if option_affect_files_checkbox.checked do
                result = cleaner.rename_file current_path new_path
            
            if result then
                cleaner.set_bitmap_path current_path new_path
            else
            (
                print("ERROR: Could not rename file: " + current_path)
                error_counter += 1
            ) --end else
        ) --end for
        
        cleaner.collect_bitmap_paths()
        refresh_bitmap_list()

        if error_counter == 0 then
            status_label.text = "INFO: Selected maps have been modified"
        else
            status_label.text = "WARNING: Modification complete, but errors occurred. Check listener for details"
    ) --end fn
    
    /**********************************************************************
    Reset modify button clicked event handler. Resets the modify fields to
    their default values
    **********************************************************************/
    fn reset_modify_button_pressed sender events =
    (
        modify_text.text = ""

        status_label.text = "INFO: Modify fields have been reset"
    ) --end fn
    
    /**********************************************************************
    Find/Replace selected clicked event handler. Performs a find/replace
    operation on the selected bitmaps
    **********************************************************************/
    fn find_replace_button_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text
        local find_mode = 1
        local message = ""
        local result = true
        local error_counter = 0
        
        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No maps selected"
            return false
        ) --end if

        if find_text.text == "" do
        (
            status_label.text = "ERROR: Find text is blank"
            return false
        ) --end if

        if (findString replace_text.text find_text.text) != undefined do
        (
            first_only_checkbox.checked = true
            message = "WARNING: When the \"Find\" text is in the \"Replace\""
            message += " text only the first occurrence can be changed. "
        ) --end if

        case of
        (
            (mode_filename_radio.checked): find_mode = 2
            (mode_extension_radio.checked): find_mode = 3
            (mode_all_radio.checked): find_mode = 4
        ) --end case
                
        status_label.text = "INFO: Editing selected maps. Please wait..."
        if ((maxVersion())[1] / 1000) > 12 do windows.processPostedMessages()
        
        for current_path in bitmap_paths do
        (
            new_path = cleaner.get_find_replace_result current_path find_mode \
                                                             find_text.text replace_text.text \
                                                             match_case_checkbox.checked \
                                                             first_only_checkbox.checked
            
            if option_affect_files_checkbox.checked do
                result = cleaner.rename_file current_path new_path
            
            if result then
                cleaner.set_bitmap_path current_path new_path
            else
            (
                print("ERROR: Could not rename file: " + current_path)
                error_counter += 1
            ) --end else
        ) --end for
        
        cleaner.collect_bitmap_paths()
        refresh_bitmap_list()

        if error_counter == 0 then
        (
            message += "INFO: Replaced \"" + find_text.text + "\" with \""
            message += replace_text.text + "\" in selected maps"
        ) --end if
        else
        (
            message += "WARNING: Find/Replace complete, but errors occurred."
            message += " Check listener for details"
        ) --end else
        
         status_label.text = message   
    ) --end fn
    
    /**********************************************************************
    Reset find button clicked event handler. Resets the find/replace fields
    to their default values
    **********************************************************************/
    fn reset_find_button_pressed sender events =
    (
        find_text.text = ""
        replace_text.text = ""
        match_case_checkbox.Checked = false
        first_only_checkbox.Checked = false

        status_label.text = "INFO: Find/Replace fields have been reset"
    ) --end fn
    
    /**********************************************************************
    Select missing map clicked event handler. Highlights all missing maps
    **********************************************************************/
    fn missing_map_button_pressed sender events =
    (
        local missing_maps = cleaner.get_missing_maps()
        
        for i = 0 to (bitmap_list.Items.Count - 1) do
        (
            found = findItem missing_maps (bitmap_list.Items.Item i).text
            
            if found > 0 then
                (bitmap_list.Items.Item i).Selected = true
            else
                (bitmap_list.Items.Item i).Selected = false
        ) --end for

        bitmap_list.Select()
        
        message = (missing_maps.count as string) + " missing maps found"
        if missing_maps.count > 0 then
            message = "WARNING: " + message
        else
            message = "INFO: " + message
        status_label.text = message
    ) --end fn
    
    /**********************************************************************
    Export maps clicked event handler. Copies selected bitmaps to the user
    selected directory
    **********************************************************************/
    fn map_export_button_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text
        local update_scene = false
        local error_counter = 0
        local message = "INFO: "
        
        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No paths selected"
            return false
        ) --end if
        
        local export_path = getSavePath caption:"Export Directory"

        if export_path == undefined do
        (
            status_label.text = "WARNING: Map export cancelled"
            return false
        ) --end if

        if sender.name == "export_update_button" do
            update_scene = true
        
        progress_bar.Value = 0
        progress_bar.Visible = true
        progress_bar.Maximum = bitmap_paths.count
        
        status_label.text = "INFO: Exporting maps. Please wait..."
        if ((maxVersion())[1] / 1000) > 12 do windows.processPostedMessages()
        
        for i = 1 to bitmap_paths.count do
        (
            result = cleaner.export_maps export_path bitmap_paths[i] update_scene
            
            if not result do
                error_counter += 1
            
            progress_bar.PerformStep()
        ) --end for
        
        if update_scene do
        (
            cleaner.collect_bitmap_paths()
            refresh_bitmap_list()
        ) --end if
        
        progress_bar.Visible = false
        progress_bar.Value = 0

        if error_counter > 0 do
            message = "WARNING: "

        message += ("Maps exported. " + (error_counter as string) + " errors.")
        message += " Check listener for details"

        status_label.text = message
    ) --end fn
    
    /**********************************************************************
    Show materials clicked event handler. Displays the bitmap nodes in the slate
    material editor
    **********************************************************************/
    fn show_materials_button_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text
        local bitmap_nodes = #()
        local material_dependents = #()

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No paths selected"
            return false
        ) --end if

        if SME == undefined do
        (
            status_label.text = "ERROR: Slate material editor was not found"
            return false
        ) --end if

        for current_path in bitmap_paths do
            join bitmap_nodes (cleaner.get_nodes_using_path current_path)

        for current_node in bitmap_nodes do
            join material_dependents (cleaner.get_materials_using_node current_node)

        material_dependents = makeUniqueArray material_dependents
        
        if material_dependents.count == 0 do
        (
            status_label.text = "WARNING: Selected maps not found in any materials"
            return false
        ) --end if

        --CRITICAL! The slate editor must be open at this point!
        SME.Open()
        sme_view = SME.getView (SME.activeView)

        if sme_view == undefined do
        (
            sme_view = SME.createView "View1"
            sme_view = SME.getView (SME.activeView)
        ) --end if

        for i = 1 to material_dependents.count do
        (
            node_postition = [0, (i * 100)]
            sme_view.createNode material_dependents[i] &node_postition
        ) --end for

        try
        (
            --sme_view.zoomExtents type:#all
            hwnd = windows.getChildHWND 0 "Slate Material Editor"
            WM_COMMAND = 0x111
            ID = 0xA02C -- Layout All Button source ID
            windows.sendMessage hwnd[1] WM_COMMAND ID 0 -- press button
        ) --end try
        catch()

        status_label.text = "INFO: Showing materials using map(s) in slate editor"
    ) --end fn
    
    /**********************************************************************
    Reset viewport clicked event handler. Resets all current viewport
    backgrounds
    **********************************************************************/
    fn reset_viewport_button_pressed sender events =
    (
        for i = 1 to viewport.numViews do
        (
            viewport.activeViewport = i
            backgroundImageFileName = ""
        )--end for

        status_label.text = "INFO: Viewport backgrounds have been reset"
    ) --end fn
    
    /**********************************************************************
    Instance bitmaps clicked event handler. Instances all bitmaps using
    the same file path
    **********************************************************************/
    fn instance_map_button_pressed sender events =
    (
        local bitmap_paths = makeUniqueArray cleaner.bitmap_paths
        local total_maps = 0
        local counter = 0

        for current_path in bitmap_paths do
        (
            current_nodes = cleaner.get_nodes_using_path current_path

            if current_nodes.count > 1 do
            (
                counter += (current_nodes.count - 1)
                master_node = current_nodes[1]
                deleteItem current_nodes 1

                for replace_node in current_nodes do
                    replaceInstances replace_node master_node
            ) --end if
        ) --end for

        cleaner.collect_bitmap_nodes()
        total_maps = refresh_bitmap_list()

        bitmap_group.text = ("Maps: " + (total_maps as string) + "   Nodes: " + (cleaner.bitmap_nodes.count as string))

        status_label.text = "INFO: " + (counter as string) + " instances created"
    ) --end fn
    
    /**********************************************************************
    View map clicked event handler. Displays the selected maps in a VFB
    **********************************************************************/
    fn view_map_button_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text
        local error_counter = 0

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No paths selected"
            return false
        ) --end if

        for current_path in bitmap_paths do
        (
            full_path = cleaner.get_absolute_path current_path

            if (doesFileExist full_path) then
                display (openBitMap full_path)
            else
            (
                print("ERROR: Unable to dispaly bitmap " + full_path)
                error_counter += 1
            ) --end else
        ) --end for

        if error_counter == 0 then
            message = "INFO: Displaying bitmap(s)"
        else
            message = "WARNING: Errors occurred displaying bitmaps. Check listener"

        status_label.text = message
    ) --end fn
    
    /**********************************************************************
    Explore directory clicked event handler. Opens an explorer window
    starting at the selected maps directory
    **********************************************************************/
    fn explore_directory_button_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No paths selected"
            return false
        ) --end if

        for current_path in bitmap_paths do
        (
            full_path = cleaner.get_absolute_path current_path
            path_directory = getFilenamePath full_path

            if (doesFileExist path_directory) do
                shellLaunch "explorer.exe" path_directory
        ) --end for

        status_label.text = "INFO: Exploring map directories"
    ) --end fn
    
    /**********************************************************************
    Print dependents clicked event handler. Prints the dependents of the
    selected bitmaps
    **********************************************************************/
    fn print_dependent_button_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text
        local bitmap_nodes = #()

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No paths selected"
            return false
        ) --end if

        for current_path in bitmap_paths do
            join bitmap_nodes (cleaner.get_nodes_using_path current_path)
        
        for current in bitmap_nodes do
        (
            print("BITMAP NODE: " + current.name)
            print("BITMAP PATH: " + cleaner.get_node_path current)
            print("DEPENDENTS:")
            print(refs.dependents current)
            print("")
        ) --end for
        
        status_label.text = "INFO: Dependents printed to listener"
    ) --end fn
    
    /**********************************************************************
    Reset unwrap modifiers clicked event handler. Removes all bitmaps from
    all unwrap modifiers
    **********************************************************************/
    fn reset_unwrap_modifiers_button_pressed sender events =
	(
        cleaner.reset_unwrap_textures()
        
        status_label.text = "INFO: Unwrap modifier textures have been reset"
	)--end fn
    
    /********************************************************************
    List resize event handler. Ensures the last column will fill the
    remaining space in the list
    ********************************************************************/
    fn bitmap_list_resized sender events =
    (
        (bitmap_list.Columns.Item 2).width = -2
    ) --end fn
    
    /**********************************************************************
    Copy path menu clicked event handler. Copies the last selected bitmap path to the clipboard
    **********************************************************************/
    fn copy_path_menu_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No paths selected"
            return false
        ) --end if

        file_path = cleaner.get_file_path bitmap_paths[bitmap_paths.count]
        
        setclipboardText file_path
        
        status_label.text = "INFO: Path copied to clipboard"
    ) --end fn
    
    /**********************************************************************
    Copy filename menu clicked event handler. Copies the last selected bitmap name to the clipboard
    **********************************************************************/
    fn copy_filename_menu_pressed sender events =
    (
        local user_selection = for i = 0 to (bitmap_list.SelectedIndices.Count - 1) collect bitmap_list.SelectedIndices.Item i
        local bitmap_paths = for index in user_selection collect (bitmap_list.Items.Item index).text

        if user_selection.count == 0 do
        (
            status_label.text = "ERROR: No paths selected"
            return false
        ) --end if

        file_name = getFilenameFile bitmap_paths[bitmap_paths.count]
        
        setclipboardText file_name
        
        status_label.text = "INFO: Filename copied to clipboard"
    ) --end fn
    
    /**********************************************************************
    Select all button clicked event handler. Selects all maps in the
    bitmap listbox
    **********************************************************************/
    fn select_all_button_pressed sender events =
    (
        for i = 0 to (bitmap_list.Items.Count - 1) do
            (bitmap_list.Items.Item i).Selected = true

        bitmap_list.Select()
        
        status_label.text = "INFO: All bitmaps have been selected"
    ) --end fn
    
    /**********************************************************************
    Select invert button clicked event handler. Selects all maps in the
    bitmap listbox that are not selected and deselects selected maps
    **********************************************************************/
    fn select_invert_button_pressed sender events =
    (
        for i = 0 to (bitmap_list.Items.Count - 1) do
            (bitmap_list.Items.Item i).Selected = not ((bitmap_list.Items.Item i).Selected)

        bitmap_list.Select()

        status_label.text = "INFO: Bitmap selection has been inverted"
    ) --end fn

    /**********************************************************************
    Select none button clicked event handler. Deselects all maps in the
    bitmap listbox
    **********************************************************************/
    fn select_none_button_pressed sender events =
    (
        for i = 0 to (bitmap_list.Items.Count - 1) do
            (bitmap_list.Items.Item i).Selected = false

        bitmap_list.Select()

        status_label.text = "INFO: All bitmaps have been deselected"
    ) --end fn
    
    /********************************************************************
    Update list event handler. Refreshes the bitmap list
    ********************************************************************/
    fn update_button_pressed sender events =
    (
        cleaner.collect_bitmap_nodes()
        local total_maps = refresh_bitmap_list()

        bitmap_group.text = ("Maps: " + (total_maps as string) + "   Nodes: " + (cleaner.bitmap_nodes.count as string))
        status_label.text = ("INFO: Bitmap list updated. Total maps: " + (total_maps as string) + " Total bitmap nodes: " + (cleaner.bitmap_nodes.count as string))
    ) --end fn
    
    /********************************************************************
    UI creation
    ********************************************************************/
    fn setup_ui =
    (
        bitmap_cleaner_form = dotNetObject "MaxCustomControls.MaxForm"
        bitmap_cleaner_form.name = "bitmap_cleaner_form"
        bitmap_cleaner_form.text = "Bitmap Cleaner"
        bitmap_cleaner_form.width = 1000
        bitmap_cleaner_form.height = 450
        dotNet.setLifetimeControl bitmap_cleaner_form #dotnet
        dotNet.addEventHandler bitmap_cleaner_form "FormClosing" bitmap_cleaner_form_closing
        bitmap_cleaner_form.suspendLayout()
        
        /********************************************************************
        Form Layout
        ********************************************************************/
        local bitmap_form_layout = dotNetObject "System.Windows.Forms.TableLayoutPanel"
        bitmap_form_layout.Name = "bitmap_form_layout"
        bitmap_form_layout.Location = dotNetObject "System.Drawing.Point" 5 5
        bitmap_form_layout.Size = dotNetObject "System.Drawing.Size" 900 350
        bitmap_form_layout.Dock = bitmap_form_layout.Dock.Fill
        bitmap_form_layout.Margin = dotNetobject "System.Windows.Forms.Padding" 5 5 5 5
        bitmap_form_layout.Padding = dotNetobject "System.Windows.Forms.Padding" 5 5 5 5
        bitmap_form_layout.ColumnCount = 2
        bitmap_form_layout.ColumnStyles.Add (dotnetobject "System.Windows.Forms.ColumnStyle" (dotnetclass "System.Windows.Forms.SizeType").Absolute 350)
        bitmap_form_layout.ColumnStyles.Add (dotnetobject "System.Windows.Forms.ColumnStyle" (dotnetclass "System.Windows.Forms.SizeType").Percent 100)
        bitmap_form_layout.RowCount = 1
        bitmap_form_layout.RowStyles.Add (dotnetobject "System.Windows.Forms.RowStyle" (dotnetclass "System.Windows.Forms.SizeType").Percent 100)
        bitmap_form_layout.suspendLayout()
        
        bitmap_cleaner_form.Controls.Add bitmap_form_layout
        
        /********************************************************************
        Controls Tab
        ********************************************************************/
        local controls_tab = dotNetObject "System.Windows.Forms.TabControl"
        controls_tab.Name = "controls_tab"
        controls_tab.Location = dotNetObject "System.Drawing.Point" 8 8
        controls_tab.Size = dotNetObject "System.Drawing.Size" 344 362
        controls_tab.Dock = controls_tab.Dock.Fill
        controls_tab.SelectedIndex = 0
        controls_tab.suspendLayout()
        
        bitmap_form_layout.Controls.Add controls_tab 0 0
        
        /********************************************************************
        Search Tab Page
        ********************************************************************/
        local search_page = dotNetObject "System.Windows.Forms.TabPage"
        search_page.Name = "search_page"
        search_page.Text = "Search"
        search_page.Location = dotNetObject "System.Drawing.Point" 4 25
        search_page.Size = dotNetObject "System.Drawing.Size" 336 333
        
        dn_color = (dotNetClass "System.Drawing.Color").FromArgb
        current_color = (colorMan.getColor #background) * 255
        search_page.BackColor = dn_color current_color[1] current_color[2] current_color[3]
        
        search_page.Padding = dotNetobject "System.Windows.Forms.Padding" 5 5 5 5
        search_page.suspendLayout()
        
        controls_tab.Controls.Add search_page
        
        /********************************************************************
        Search Controls
        ********************************************************************/
        local search_directory_group = dotNetObject "System.Windows.Forms.GroupBox"
        search_directory_group.name = "search_directory_group"
        search_directory_group.text = "Directory"
        search_directory_group.Location = dotNetObject "System.Drawing.Point" 5 5
        search_directory_group.Size = dotNetObject "System.Drawing.Size" 324 75
        search_directory_group.flatStyle = search_directory_group.flatStyle.system
        search_directory_group.suspendLayout()

        search_page.Controls.Add search_directory_group

        search_directory_text = dotNetObject "System.Windows.Forms.TextBox"
        search_directory_text.Name = "search_directory_text"
        search_directory_text.Location = dotNetObject "System.Drawing.Point" 10 20
        search_directory_text.Size = dotNetObject "System.Drawing.Size" 260 22
        
        search_directory_group.Controls.Add search_directory_text
        
        local search_browse_button = dotNetObject "System.Windows.Forms.Button"
        search_browse_button.Name = "search_browse_button"
        search_browse_button.Text = "..."
        search_browse_button.Location = dotNetObject "System.Drawing.Point" 280 20
        search_browse_button.Size = dotNetObject "System.Drawing.Size" 35 25
        search_browse_button.flatStyle = search_browse_button.flatStyle.system
        dotNet.setLifetimeControl search_browse_button #dotnet
        dotNet.addEventHandler search_browse_button "Click" search_browse_button_pressed
        
        search_directory_group.Controls.Add search_browse_button
        
        include_sub_directory_checkbox = dotNetObject "System.Windows.Forms.CheckBox"
        include_sub_directory_checkbox.Name = "include_sub_directory_checkbox"
        include_sub_directory_checkbox.Text = "Include Sub Directories"
        include_sub_directory_checkbox.Location = dotNetObject "System.Drawing.Point" 10 50
        include_sub_directory_checkbox.Size = dotNetObject "System.Drawing.Size" 105 21
        include_sub_directory_checkbox.Checked = True
        include_sub_directory_checkbox.AutoSize = True
        
        search_directory_group.Controls.Add include_sub_directory_checkbox
        
        local search_directory_button = dotNetObject "System.Windows.Forms.Button"
        search_directory_button.Name = "search_directory_button"
        search_directory_button.Text = "Search for Selected Maps"
        search_directory_button.Location = dotNetObject "System.Drawing.Point" 5 90
        search_directory_button.Size = dotNetObject "System.Drawing.Size" 324 25
        search_directory_button.flatStyle = search_directory_button.flatStyle.system
        dotNet.setLifetimeControl search_directory_button #dotnet
        dotNet.addEventHandler search_directory_button "Click" search_directory_button_pressed
        
        search_page.Controls.Add search_directory_button
        
        /********************************************************************
        Edit Tab Page
        ********************************************************************/
        local edit_page = dotNetObject "System.Windows.Forms.TabPage"
        edit_page.Name = "edit_page"
        edit_page.Text = "Edit"
        edit_page.Location = dotNetObject "System.Drawing.Point" 4 25
        edit_page.Size = dotNetObject "System.Drawing.Size" 336 333
        
        dn_color = (dotNetClass "System.Drawing.Color").FromArgb
        current_color = (colorMan.getColor #background) * 255
        edit_page.BackColor = dn_color current_color[1] current_color[2] current_color[3]
        
        edit_page.Padding = dotNetobject "System.Windows.Forms.Padding" 5 5 5 5
        edit_page.suspendLayout()
        
        controls_tab.Controls.Add edit_page
        
        /********************************************************************
        Edit Controls
        ********************************************************************/
        local mode_settings_group = dotNetObject "System.Windows.Forms.GroupBox"
        mode_settings_group.name = "mode_settings_group"
        mode_settings_group.text = "Mode"
        mode_settings_group.Location = dotNetObject "System.Drawing.Point" 5 5
        mode_settings_group.Size = dotNetObject "System.Drawing.Size" 324 45
        mode_settings_group.flatStyle = mode_settings_group.flatStyle.system
        mode_settings_group.suspendLayout()

        edit_page.Controls.Add mode_settings_group
        
        /********************************************************************
        Mode Controls
        ********************************************************************/
        mode_path_radio = dotNetObject "System.Windows.Forms.RadioButton"
        mode_path_radio.Name = "mode_path_radio"
        mode_path_radio.Text = "Path"
        mode_path_radio.Location = dotNetObject "System.Drawing.Point" 10 20
        mode_path_radio.Size = dotNetObject "System.Drawing.Size" 60 21
        mode_path_radio.AutoSize = True
        
        mode_settings_group.Controls.Add mode_path_radio
        
        mode_filename_radio = dotNetObject "System.Windows.Forms.RadioButton"
        mode_filename_radio.Name = "mode_filename_radio"
        mode_filename_radio.Text = "Filename"
        mode_filename_radio.Location = dotNetObject "System.Drawing.Point" 75 20
        mode_filename_radio.Size = dotNetObject "System.Drawing.Size" 60 21
        mode_filename_radio.AutoSize = True
        mode_filename_radio.Checked = True
        
        mode_settings_group.Controls.Add mode_filename_radio
        
        mode_extension_radio = dotNetObject "System.Windows.Forms.RadioButton"
        mode_extension_radio.Name = "mode_extension_radio"
        mode_extension_radio.Text = "Extension"
        mode_extension_radio.Location = dotNetObject "System.Drawing.Point" 165 20
        mode_extension_radio.Size = dotNetObject "System.Drawing.Size" 60 21
        mode_extension_radio.AutoSize = True
        
        mode_settings_group.Controls.Add mode_extension_radio
        
        mode_all_radio = dotNetObject "System.Windows.Forms.RadioButton"
        mode_all_radio.Name = "mode_all_radio"
        mode_all_radio.Text = "All"
        mode_all_radio.Location = dotNetObject "System.Drawing.Point" 260 20
        mode_all_radio.Size = dotNetObject "System.Drawing.Size" 60 21
        mode_all_radio.AutoSize = True
        
        mode_settings_group.Controls.Add mode_all_radio
        
        /********************************************************************
        Option Controls
        ********************************************************************/
        local option_settings_group = dotNetObject "System.Windows.Forms.GroupBox"
        option_settings_group.name = "option_settings_group"
        option_settings_group.text = "Options"
        option_settings_group.Location = dotNetObject "System.Drawing.Point" 5 55
        option_settings_group.Size = dotNetObject "System.Drawing.Size" 324 45
        option_settings_group.flatStyle = option_settings_group.flatStyle.system
        option_settings_group.suspendLayout()

        edit_page.Controls.Add option_settings_group
        
        option_affect_files_checkbox = dotNetObject "System.Windows.Forms.CheckBox"
        option_affect_files_checkbox.Name = "option_affect_files_checkbox"
        option_affect_files_checkbox.Text = "Affect Files When Editing (use with caution)"
        option_affect_files_checkbox.Location = dotNetObject "System.Drawing.Point" 10 20
        option_affect_files_checkbox.Size = dotNetObject "System.Drawing.Size" 105 21
        option_affect_files_checkbox.AutoSize = True
        
        option_settings_group.Controls.Add option_affect_files_checkbox
        
        /********************************************************************
        Modify Controls
        ********************************************************************/
        local modify_settings_group = dotNetObject "System.Windows.Forms.GroupBox"
        modify_settings_group.name = "modify_settings_group"
        modify_settings_group.text = "Modify"
        modify_settings_group.Location = dotNetObject "System.Drawing.Point" 5 105
        modify_settings_group.Size = dotNetObject "System.Drawing.Size" 324 80
        modify_settings_group.flatStyle = modify_settings_group.flatStyle.system
        modify_settings_group.suspendLayout()

        edit_page.Controls.Add modify_settings_group

        modify_text = dotNetObject "System.Windows.Forms.TextBox"
        modify_text.Name = "modify_text"
        modify_text.Location = dotNetObject "System.Drawing.Point" 10 20
        modify_text.Size = dotNetObject "System.Drawing.Size" 304 22
        
        modify_settings_group.Controls.Add modify_text
        
        local modify_paths_button = dotNetObject "System.Windows.Forms.Button"
        modify_paths_button.Name = "modify_paths_button"
        modify_paths_button.Text = "Modify Selected Maps"
        modify_paths_button.Location = dotNetObject "System.Drawing.Point" 10 50
        modify_paths_button.Size = dotNetObject "System.Drawing.Size" 245 25
        modify_paths_button.flatStyle = modify_paths_button.flatStyle.system
        dotNet.setLifetimeControl modify_paths_button #dotnet
        dotNet.addEventHandler modify_paths_button "Click" modify_path_button_pressed
        
        modify_settings_group.Controls.Add modify_paths_button
        
        local modify_reset_button = dotNetObject "System.Windows.Forms.Button"
        modify_reset_button.Text = "Reset"
        modify_reset_button.Name = "modify_reset_button"
        modify_reset_button.Location = dotNetObject "System.Drawing.Point" 265 50
        modify_reset_button.Size = dotNetObject "System.Drawing.Size" 50 25
        modify_reset_button.flatStyle = modify_reset_button.flatStyle.system
        dotNet.setLifetimeControl modify_reset_button #dotnet
        dotNet.addEventHandler modify_reset_button "Click" reset_modify_button_pressed
        
        modify_settings_group.Controls.Add modify_reset_button

        /********************************************************************
        Find/Replace Controls
        ********************************************************************/
        local find_replace_group = dotNetObject "System.Windows.Forms.GroupBox"
        find_replace_group.name = "find_replace_group"
        find_replace_group.text = "Find/Replace"
        find_replace_group.Location = dotNetObject "System.Drawing.Point" 5 190
        find_replace_group.Size = dotNetObject "System.Drawing.Size" 324 160
        find_replace_group.flatStyle = find_replace_group.flatStyle.system
        find_replace_group.suspendLayout()
        
        edit_page.Controls.Add find_replace_group
                
        local find_label = dotNetObject "System.Windows.Forms.Label"
        find_label.Name = "find_label"
        find_label.Text = "Find:"
        find_label.Location = dotNetObject "System.Drawing.Point" 10 22
        find_label.Size = dotNetObject "System.Drawing.Size" 70 17
        find_label.AutoSize = True
        
        find_replace_group.Controls.Add find_label
        
        find_text = dotNetObject "System.Windows.Forms.TextBox"
        find_text.Name = "find_text"
        find_text.Location = dotNetObject "System.Drawing.Point" 85 20
        find_text.Size = dotNetObject "System.Drawing.Size" 225 22
        
        find_replace_group.Controls.Add find_text
        
        local replace_label = dotNetObject "System.Windows.Forms.Label"
        replace_label.Name = "replace_label"
        replace_label.Text = "Replace:"
        replace_label.Location = dotNetObject "System.Drawing.Point" 10 52
        replace_label.Size = dotNetObject "System.Drawing.Size" 70 17
        replace_label.AutoSize = True
        
        find_replace_group.Controls.Add replace_label
        
        replace_text = dotNetObject "System.Windows.Forms.TextBox"
        replace_text.Name = "replace_text"
        replace_text.Location = dotNetObject "System.Drawing.Point" 85 50
        replace_text.Size = dotNetObject "System.Drawing.Size" 225 22
        
        find_replace_group.Controls.Add replace_text
        
        match_case_checkbox = dotNetObject "System.Windows.Forms.CheckBox"
        match_case_checkbox.Name = "match_case_checkbox"
        match_case_checkbox.Text = "Match Case"
        match_case_checkbox.Location = dotNetObject "System.Drawing.Point" 10 80
        match_case_checkbox.Size = dotNetObject "System.Drawing.Size" 105 21
        match_case_checkbox.AutoSize = True
        
        find_replace_group.Controls.Add match_case_checkbox

        first_only_checkbox = dotNetObject "System.Windows.Forms.CheckBox"
        first_only_checkbox.Name = "first_only_checkbox"
        first_only_checkbox.Text = "First Occurence Only"
        first_only_checkbox.Location = dotNetObject "System.Drawing.Point" 10 105
        first_only_checkbox.Size = dotNetObject "System.Drawing.Size" 165 21
        first_only_checkbox.AutoSize = True
            
        find_replace_group.Controls.Add first_only_checkbox
        
        local find_replace_button = dotNetObject "System.Windows.Forms.Button"
        find_replace_button.Name = "find_replace_button"
        find_replace_button.Text = "Find/Replace Selected Maps"
        find_replace_button.Location = dotNetObject "System.Drawing.Point" 10 130
        find_replace_button.Size = dotNetObject "System.Drawing.Size" 245 25
        find_replace_button.flatStyle = find_replace_button.flatStyle.system
        dotNet.setLifetimeControl find_replace_button #dotnet
        dotNet.addEventHandler find_replace_button "Click" find_replace_button_pressed
        
        find_replace_group.Controls.Add find_replace_button
        
        local find_reset_button = dotNetObject "System.Windows.Forms.Button"
        find_reset_button.Text = "Reset"
        find_reset_button.Name = "find_reset_button"
        find_reset_button.Location = dotNetObject "System.Drawing.Point" 265 130
        find_reset_button.Size = dotNetObject "System.Drawing.Size" 50 25
        find_reset_button.flatStyle = find_reset_button.flatStyle.system
        dotNet.setLifetimeControl find_reset_button #dotnet
        dotNet.addEventHandler find_reset_button "Click" reset_find_button_pressed
        
        find_replace_group.Controls.Add find_reset_button

        /********************************************************************
        Misc Tab Page
        ********************************************************************/
        local misc_page = dotNetObject "System.Windows.Forms.TabPage"
        misc_page.Name = "misc_page"
        misc_page.Text = "Misc"
        misc_page.Location = dotNetObject "System.Drawing.Point" 4 25
        misc_page.Size = dotNetObject "System.Drawing.Size" 336 588
            
        dn_color = (dotNetClass "System.Drawing.Color").FromArgb
        current_color = (colorMan.getColor #background) * 255
        misc_page.BackColor = dn_color current_color[1] current_color[2] current_color[3]
        
        misc_page.Padding = dotNetobject "System.Windows.Forms.Padding" 5 5 5 5
        misc_page.suspendLayout()
        
        controls_tab.Controls.Add misc_page
        
        /********************************************************************
        Misc Controls
        ********************************************************************/
        local select_missing_button = dotNetObject "System.Windows.Forms.Button"
        select_missing_button.Name = "select_missing_button"
        select_missing_button.Text = "Select Missing Maps"
        select_missing_button.Location = dotNetObject "System.Drawing.Point" 6 6
        select_missing_button.Size = dotNetObject "System.Drawing.Size" 324 25
        select_missing_button.flatStyle = select_missing_button.flatStyle.system
        select_missing_button.Anchor = dotnet.combineEnums select_missing_button.anchor.Top select_missing_button.anchor.Left select_missing_button.anchor.Right
        dotNet.setLifetimeControl select_missing_button #dotnet
        dotNet.addEventHandler select_missing_button "Click" missing_map_button_pressed

        misc_page.Controls.Add select_missing_button
        
        local export_map_button = dotNetObject "System.Windows.Forms.Button"
        export_map_button.Name = "export_map_button"
        export_map_button.Text = "Export Selected Maps"
        export_map_button.Location = dotNetObject "System.Drawing.Point" 6 35
        export_map_button.Size = dotNetObject "System.Drawing.Size" 324 25
        export_map_button.flatStyle = export_map_button.flatStyle.system
        export_map_button.Anchor = dotnet.combineEnums export_map_button.anchor.Top export_map_button.anchor.Left export_map_button.anchor.Right
        dotNet.setLifetimeControl export_map_button #dotnet
        dotNet.addEventHandler export_map_button "Click" map_export_button_pressed

        misc_page.Controls.Add export_map_button

        local export_update_button = dotNetObject "System.Windows.Forms.Button"
        export_update_button.Name = "export_update_button"
        export_update_button.Text = "Export Selected Maps and Update Scene Paths"
        export_update_button.Location = dotNetObject "System.Drawing.Point" 6 64
        export_update_button.Size = dotNetObject "System.Drawing.Size" 324 25
        export_update_button.flatStyle = export_update_button.flatStyle.system
        export_update_button.Anchor = dotnet.combineEnums export_update_button.anchor.Top export_update_button.anchor.Left export_update_button.anchor.Right
        dotNet.setLifetimeControl export_update_button #dotnet
        dotNet.addEventHandler export_update_button "Click" map_export_button_pressed

        misc_page.Controls.Add export_update_button
        
        local show_materials_button = dotNetObject "System.Windows.Forms.Button"
        show_materials_button.Name = "show_materials_button"
        show_materials_button.Text = "Show Materials Using Maps"
        show_materials_button.Location = dotNetObject "System.Drawing.Point" 6 93
        show_materials_button.Size = dotNetObject "System.Drawing.Size" 324 25
        show_materials_button.flatStyle = show_materials_button.flatStyle.system
        show_materials_button.Anchor = dotnet.combineEnums show_materials_button.anchor.Top show_materials_button.anchor.Left show_materials_button.anchor.Right
        dotNet.setLifetimeControl show_materials_button #dotnet
        dotNet.addEventHandler show_materials_button "Click" show_materials_button_pressed

        misc_page.Controls.Add show_materials_button
        
        local reset_viewport_button = dotNetObject "System.Windows.Forms.Button"
        reset_viewport_button.Name = "reset_viewport_button"
        reset_viewport_button.Text = "Reset Viewport Backgrounds"
        reset_viewport_button.Location = dotNetObject "System.Drawing.Point" 6 122
        reset_viewport_button.Size = dotNetObject "System.Drawing.Size" 324 25
        reset_viewport_button.flatStyle = reset_viewport_button.flatStyle.system
        reset_viewport_button.Anchor = dotnet.combineEnums reset_viewport_button.anchor.Top reset_viewport_button.anchor.Left reset_viewport_button.anchor.Right
        dotNet.setLifetimeControl reset_viewport_button #dotnet
        dotNet.addEventHandler reset_viewport_button "Click" reset_viewport_button_pressed

        misc_page.Controls.Add reset_viewport_button
        
        local instance_map_button = dotNetObject "System.Windows.Forms.Button"
        instance_map_button.Name = "instance_map_button"
        instance_map_button.Text = "Instance Bitmaps Using Same Path"
        instance_map_button.Location = dotNetObject "System.Drawing.Point" 6 151
        instance_map_button.Size = dotNetObject "System.Drawing.Size" 324 25
        instance_map_button.flatStyle = instance_map_button.flatStyle.system
        instance_map_button.Anchor = dotnet.combineEnums instance_map_button.anchor.Top instance_map_button.anchor.Left instance_map_button.anchor.Right
        dotNet.setLifetimeControl instance_map_button #dotnet
        dotNet.addEventHandler instance_map_button "Click" instance_map_button_pressed

        misc_page.Controls.Add instance_map_button
        
        local view_map_button = dotNetObject "System.Windows.Forms.Button"
        view_map_button.Name = "view_map_button"
        view_map_button.Text = "View Maps"
        view_map_button.Location = dotNetObject "System.Drawing.Point" 6 180
        view_map_button.Size = dotNetObject "System.Drawing.Size" 324 25
        view_map_button.flatStyle = view_map_button.flatStyle.system
        view_map_button.Anchor = dotnet.combineEnums view_map_button.anchor.Top view_map_button.anchor.Left view_map_button.anchor.Right
        dotNet.setLifetimeControl view_map_button #dotnet
        dotNet.addEventHandler view_map_button "Click" view_map_button_pressed

        misc_page.Controls.Add view_map_button
        
        local explore_directory_button = dotNetObject "System.Windows.Forms.Button"
        explore_directory_button.Name = "explore_directory_button"
        explore_directory_button.Text = "Explore Map Directory"
        explore_directory_button.Location = dotNetObject "System.Drawing.Point" 6 209
        explore_directory_button.Size = dotNetObject "System.Drawing.Size" 324 25
        explore_directory_button.flatStyle = explore_directory_button.flatStyle.system
        explore_directory_button.Anchor = dotnet.combineEnums explore_directory_button.anchor.Top explore_directory_button.anchor.Left explore_directory_button.anchor.Right
        dotNet.setLifetimeControl explore_directory_button #dotnet
        dotNet.addEventHandler explore_directory_button "Click" explore_directory_button_pressed

        misc_page.Controls.Add explore_directory_button
        
        local print_dependent_button = dotNetObject "System.Windows.Forms.Button"
        print_dependent_button.Name = "print_dependent_button"
        print_dependent_button.Text = "Print Map Dependents"
        print_dependent_button.Location = dotNetObject "System.Drawing.Point" 6 238
        print_dependent_button.Size = dotNetObject "System.Drawing.Size" 324 25
        print_dependent_button.flatStyle = print_dependent_button.flatStyle.system
        print_dependent_button.Anchor = dotnet.combineEnums print_dependent_button.anchor.Top print_dependent_button.anchor.Left print_dependent_button.anchor.Right
        dotNet.setLifetimeControl print_dependent_button #dotnet
        dotNet.addEventHandler print_dependent_button "Click" print_dependent_button_pressed

        misc_page.Controls.Add print_dependent_button
        
        local reset_unwrap_modifiers_button = dotNetObject "System.Windows.Forms.Button"
        reset_unwrap_modifiers_button.Name = "reset_unwrap_modifiers_button"
        reset_unwrap_modifiers_button.Text = "Reset Unwrap Modifier Textures"
        reset_unwrap_modifiers_button.Location = dotNetObject "System.Drawing.Point" 6 267
        reset_unwrap_modifiers_button.Size = dotNetObject "System.Drawing.Size" 324 25
        reset_unwrap_modifiers_button.flatStyle = reset_unwrap_modifiers_button.flatStyle.system
        reset_unwrap_modifiers_button.Anchor = dotnet.combineEnums reset_unwrap_modifiers_button.anchor.Top reset_unwrap_modifiers_button.anchor.Left reset_unwrap_modifiers_button.anchor.Right
        dotNet.setLifetimeControl reset_unwrap_modifiers_button #dotnet
        dotNet.addEventHandler reset_unwrap_modifiers_button "Click" reset_unwrap_modifiers_button_pressed

        misc_page.Controls.Add reset_unwrap_modifiers_button
        
        /********************************************************************
        Bitmap List Controls
        ********************************************************************/
        bitmap_group = dotNetObject "System.Windows.Forms.GroupBox"    
        bitmap_group.Name = "bitmap_group"
        bitmap_group.Text = "Maps: 0 (Nodes:0)"
        bitmap_group.Location = dotNetobject "System.Drawing.Point" 355 5
        bitmap_group.Size = dotNetobject "System.Drawing.Size" 622 368
        bitmap_group.Margin = dotNetobject "System.Windows.Forms.Padding" 5 5 5 5
        bitmap_group.Padding = dotNetobject "System.Windows.Forms.Padding" 5 5 5 5
        bitmap_group.flatStyle = bitmap_group.flatStyle.system
        bitmap_group.Dock = bitmap_group.Dock.Fill
        
        bitmap_form_layout.Controls.Add bitmap_group 1 0
        
        bitmap_list = dotNetObject "System.Windows.Forms.ListView"
        bitmap_list.Name = "bitmap_list"
        bitmap_list.Size = dotNetObject "System.Drawing.Size" 607 308
        bitmap_list.Location = dotNetObject "System.Drawing.Point" 8 23
        bitmap_list.View = (dotNetClass "System.Windows.Forms.View").Details
        bitmap_list.FullRowSelect = true
        bitmap_list.HideSelection = false
        bitmap_list.Columns.Add "Paths" 50
        bitmap_list.Columns.Add "Length" 50
        bitmap_list.Columns.Add "Status" 50
        bitmap_list.Anchor = dotnet.combineEnums bitmap_list.anchor.Top bitmap_list.anchor.Left bitmap_list.anchor.Bottom bitmap_list.anchor.Right
        dotNet.setLifetimeControl bitmap_list #dotnet
        dotNet.addEventHandler bitmap_list "Resize" bitmap_list_resized
        
        dn_color = (dotNetClass "System.Drawing.Color").FromArgb
        current_color = (colorMan.getColor #window) * 255
        bitmap_list.BackColor = dn_color current_color[1] current_color[2] current_color[3]
        
        bitmap_group.Controls.Add bitmap_list
        
        /********************************************************************
        Right click menu
        ********************************************************************/
        local context_menu = dotNetObject "System.Windows.Forms.ContextMenu"

        bitmap_list.contextMenu = context_menu

        local copy_path_menu = dotNetObject "System.Windows.Forms.MenuItem"
        copy_path_menu.text = "Copy Path to Clipboard"
        dotNet.setLifetimeControl copy_path_menu #dotnet
        dotNet.addEventHandler copy_path_menu "Click" copy_path_menu_pressed

        context_menu.menuItems.add copy_path_menu

        local copy_filename_menu = dotNetObject "System.Windows.Forms.MenuItem"
        copy_filename_menu.text = "Copy File Name to Clipboard"
        dotNet.setLifetimeControl copy_filename_menu #dotnet
        dotNet.addEventHandler copy_filename_menu "Click" copy_filename_menu_pressed

        context_menu.menuItems.add copy_filename_menu

        /********************************************************************
        List Select Buttons
        ********************************************************************/
        local select_all_button = dotNetObject "System.Windows.Forms.Button"
        select_all_button.Name = "select_all_button"
        select_all_button.Text = "Select All"
        select_all_button.Location = dotNetObject "System.Drawing.Point" 8 338
        select_all_button.Size = dotNetObject "System.Drawing.Size" 125 25
        select_all_button.flatStyle = select_all_button.flatStyle.system
        select_all_button.Anchor = dotnet.combineEnums select_all_button.anchor.Left select_all_button.anchor.Bottom
        dotNet.setLifetimeControl select_all_button #dotnet
        dotNet.addEventHandler select_all_button "Click" select_all_button_pressed
        
        bitmap_group.Controls.Add select_all_button
        
        local select_invert_button = dotNetObject "System.Windows.Forms.Button"
        select_invert_button.Name = "select_invert_button"
        select_invert_button.Text = "Invert Selection"
        select_invert_button.Location = dotNetObject "System.Drawing.Point" 139 338
        select_invert_button.Size = dotNetObject "System.Drawing.Size" 125 25
        select_invert_button.flatStyle = select_invert_button.flatStyle.system
        select_invert_button.Anchor = dotnet.combineEnums select_invert_button.anchor.Left select_invert_button.anchor.Bottom
        dotNet.setLifetimeControl select_invert_button #dotnet
        dotNet.addEventHandler select_invert_button "Click" select_invert_button_pressed
        
        bitmap_group.Controls.Add select_invert_button
        
        local select_none_button = dotNetObject "System.Windows.Forms.Button"
        select_none_button.Name = "select_none_button"
        select_none_button.Text = "Select None"
        select_none_button.Location = dotNetObject "System.Drawing.Point" 270 338
        select_none_button.Size = dotNetObject "System.Drawing.Size" 125 25
        select_none_button.flatStyle = select_none_button.flatStyle.system
        select_none_button.Anchor = dotnet.combineEnums select_none_button.anchor.Left select_none_button.anchor.Bottom
        dotNet.setLifetimeControl select_none_button #dotnet
        dotNet.addEventHandler select_none_button "Click" select_none_button_pressed
        
        bitmap_group.Controls.Add select_none_button
        
        local update_list_button = dotNetObject "System.Windows.Forms.Button"
        update_list_button.Name = "update_list_button"
        update_list_button.Text = "Update"
        update_list_button.Location = dotNetObject "System.Drawing.Point" 514 338
        update_list_button.Size = dotNetObject "System.Drawing.Size" 100 25
        update_list_button.flatStyle = update_list_button.flatStyle.system
        update_list_button.Anchor = dotnet.combineEnums update_list_button.anchor.Right update_list_button.anchor.Bottom
        dotNet.setLifetimeControl update_list_button #dotnet
        dotNet.addEventHandler update_list_button "Click" update_button_pressed
        
        bitmap_group.Controls.Add update_list_button
        
        /********************************************************************
        Status bar
        ********************************************************************/
        local status_bar = dotNetObject "System.Windows.Forms.StatusStrip"
        status_bar.name = "status_bar"
        status_bar.Location = dotNetObject "System.Drawing.Point" 0 378
        status_bar.Size = dotNetObject "System.Drawing.Size" 982 25
        status_bar.dock = status_bar.dock.bottom
        status_bar.sizingGrip = false
        
        dn_color = (dotNetClass "System.Drawing.Color").FromArgb
        current_color = (colorMan.getColor #window) * 255
        status_bar.backColor = dn_color current_color[1] current_color[2] current_color[3]

        bitmap_cleaner_form.controls.add status_bar
        
        progress_bar = dotNetObject "System.Windows.Forms.ToolStripProgressBar"
        progress_bar.name = "progress_bar"
        progress_bar.Size = dotNetObject "System.Drawing.Size" 100 16
        progress_bar.Minimum = 0
        progress_bar.Maximum = 100
        progress_bar.Step = 1
        progress_bar.Visible = false
        
        status_bar.items.add progress_bar

        status_label = dotNetObject "System.Windows.Forms.ToolStripStatusLabel"
        status_label.name = "status_label"
        status_label.text = "INFO: Standing by..."
        status_label.autoSize = true

        status_bar.items.add status_label
        
        /********************************************************************
        Finalize UI
        ********************************************************************/
        find_replace_group.resumeLayout()
        modify_settings_group.resumeLayout()
        option_settings_group.resumeLayout()
        mode_settings_group.resumeLayout()
        edit_page.resumeLayout()
        search_directory_group.resumeLayout()
        search_page.resumeLayout()
        controls_tab.resumeLayout()
        bitmap_form_layout.resumeLayout()
        bitmap_cleaner_form.resumeLayout()
        
        /********************************************************************
        Refresh and Open UI
        ********************************************************************/
        cleaner.collect_bitmap_nodes()
        local total_maps = refresh_bitmap_list()

        bitmap_group.text = ("Maps: " + (total_maps as string) + "   Nodes: " + (cleaner.bitmap_nodes.count as string))
        status_label.text = ("INFO: Total maps: " + (total_maps as string) + " Total bitmap nodes: " + (cleaner.bitmap_nodes.count as string))
        
        form_position = execute (getIniSetting bitmap_cleaner_ini "Dialog" "Position")
        form_size = execute (getIniSetting bitmap_cleaner_ini "Dialog" "Size")

        if form_position == OK do
            form_position = point2 100 100
        if form_size == OK do
            form_size = point2 1000 450

        if form_position.x > sysinfo.desktopsize.x - form_size.x do
            form_position.x = sysinfo.desktopsize.x - form_size.x
        if form_position.y > sysinfo.desktopsize.y - (form_size.y + 60) do
            form_position.y = sysinfo.desktopsize.y - (form_size.y + 60)
        if form_position.x < 0 do
            form_position.x = 0
        if form_position.y < 0 do
            form_position.y = 0
        
        bitmap_cleaner_form.startPosition = bitmap_cleaner_form.startPosition.manual
        bitmap_cleaner_form.location = dotNetObject "System.Drawing.Point" form_position.x form_position.y
        bitmap_cleaner_form.size = dotNetObject "System.Drawing.Size" form_size.x form_size.y
        
        local path_column_size = bitmap_list.width - 100
        if path_column_size < 50 do path_column_size = 100    
        (bitmap_list.Columns.Item 0).Width = path_column_size

        local the_pointer = DotNetObject "System.IntPtr" (windows.getMAXHWND())
        --CRITICAL! Do not use Win32HandleWrapper, it was removed in >= 2015
        --local the_handle = DotNetObject "MaxCustomControls.Win32HandleWrapper" the_pointer
        local the_handle = (dotnetclass "System.Windows.Forms.Control").fromHandle the_pointer

        bitmap_cleaner_form.Show(the_handle)
    ) --end fn
    
    --CRITICAL! The following is required since .net/mxs can crash building a UI
    attempts = 1
    while attempts < 10 do
    (
        try
        (
            setup_ui()
            attempts = 100
        ) --end try
        catch
        (
            print("WARNING: Failed to launch Bitmap Cleaner. Attempt: " + (attempts as string))
            attempts += 1
        ) --end catch
    ) --end while
    
    if attempts == 10 do
        messageBox "ERROR: Could not launch Bitmap Cleaner" title:"Launch Error"
) --end local